/* urbcsp.cpp -- generates uniform random binary constraint satisfaction problems
*/
#include <cmath>
#include <cstdlib>
#include <iostream>
using std::cout;
using std::cerr;
#include <fstream>
using std::ofstream;
#include <string>
using std::string;
#include <sstream>
using std::stringstream;
#include <map>
using std::map;
#include <list>
using std::list;
using std::pair;

/* function declarations */
float ran2(long *idum);
void StartCSP(int N, int K, int instance, int C, stringstream &xmlOutput);
void EndCSP(stringstream &xmlOutput);
void AddConstraint(int index, int var1, int var2, stringstream &xmlOutput);
void AddNogood(int val1, int val2, stringstream &xmlOutput);
void addRelHead(int index, int t, int s, stringstream &xmlOutput);
void addTupleEnd(stringstream &xmlOutput);
void addRelTail(stringstream &xmlOutput);
// :xmlOutput will be filled with XML markup on successful return;
// Likewise, :edges will contain pairs of connections between variables.
int MakeURBCSP(int N, int D, int C, int T, long *Seed, int s, list <pair <int,int> >  &edges, stringstream &xmlOutput);
// Returns true if the graph described by this list of edges is connected, else false.
bool CheckConnectedness(int numVariables, list <pair <int,int> > &edges);

/*********************************************************************
  This file has 5 parts:
  0. This introduction.
  1. A main() function, which can be used to demonstrate MakeURBCSP().
  2. MakeURBCSP().
  3. ran2(), a random number generator.
  4. The four functions StartCSP(), AddConstraint(), AddNogood(), and 
  EndCSP(), which are called by MakeURBCSP().  The versions
  of these functions given here print out each instance, listing
  the incompatible value pairs of each constraint.  You will need
  to replace these functions with versions that mesh with your
  system and data structures.
 *********************************************************************/


/*********************************************************************
  1. A simple main() function which reads in command line parameters
  and generates CSPs.
 *********************************************************************/
int order = 0;
int main(int argc, char* argv[])
{
  int N, D, C, T, I, i, s;
	long S;

	// If too few or too many arguments, output the usage info.
	if (argc < 7 || argc > 10)
	{
		printf("usage: urbcsp #vars #vals #constraints #nogoods/supports seed instances [-s[supports]] [-f [file_tag]]\n");
		return 0;
	}

	N = atoi(argv[1]); // N: Number of variables
	D = atoi(argv[2]); // D: Number of values in each domain
	C = atoi(argv[3]); // C: Number of constraints
	T = atoi(argv[4]); // T: Number of restrictions per constraint
	S = atoi(argv[5]); // S: Random seed
	I = atoi(argv[6]); // I: Number of instances
	s = 0;

	bool outputToFile = false;
	string fileTag = string(argv[0]) + "_output_";

	// Check optional parameters
	if ( argc >= 8 ) 
	{
		int fileTagPos = 7;

		// Make supports instead of nogoods
		if ( argv[7][0] == '-' && argv[7][1] == 's' ) 
	    {
			s = 1;
			fileTagPos++; // if the -f flag exists, it's in the next slot now
		}

		// Check for the fileTag flag
		if ( argc >= fileTagPos + 1 && string(argv[fileTagPos]) == "-f")
		{
			// If we've been given a file tag as well as the flag, then
			// record the desired tag.  Otherwise, the program name is used
			// as a base tag.
			outputToFile = true;
			if (argc == fileTagPos + 2)
				fileTag = string(argv[fileTagPos+1]);
		}
	}

	if (C < N-1)
	{
		cout << "Too few constraints designated for the number of variables; ";
		cout << "At least N-1 constraints for N variables is necessary to create a connected graph.\n\n";
		cout << "Aborting.\n";
		return 0;
	}

	/* Seed passed to ran2() must initially be negative. */
	if (S > 0)
		S = -S;

	stringstream delayedOutput;

	
	for (i=0; i<I; ++i)
	{
		// Print the first run of data in MakeURBCSP
		order = 0;

        list < pair<int,int> > junk;
		if (!MakeURBCSP(N, D, C, T, &S, s, junk, delayedOutput))
			return 0;

		// Print the second run of data in MakeURBCSP
		order = 1;

        list < pair<int,int> > edges;
		if (!MakeURBCSP(N, D, C, T, &S, s, edges, delayedOutput))
			return 0;

		// If this graph isn't conneected:
		if (!CheckConnectedness(N, edges))
		{
			delayedOutput.str(""); // clear output
			--i; // try again
			continue;
		}

		// Output everything that was intended to be output during the creation
		// of the CSP.
		if (outputToFile) // To file
		{
			// stringstream is used to convert numbers to & from ASCII.
			stringstream converter;
			converter << i;
			ofstream fileOut((fileTag + converter.str() + ".xcsp").c_str());
			fileOut << delayedOutput.str();
		}
		else // To screen
			cout << delayedOutput.str() << "\n\n";

		// Now we clear it.
		delayedOutput.str("");
	}

	// If everything was printed to file, let the user know that everything
	// went smoothly and where the results are.
	if (outputToFile)
	{
		cout << "Execution completed successfully.  The resultant file(s) reside"
				<< " in the current directory if a tag was specified.  "
				<< " Otherwise, they will reside in the program directory.\n";
	}

	return 1;
}


/*********************************************************************
  2. MakeURBCSP() creates a uniform binary constraint satisfaction
  problem with a specified number of variables, domain size, 
  tightness, and number of constraints.  MakeURBCSP() calls
  four functions, StartCSP(), AddConstraint(), AddNogood(), and 
  EndCSP(), which actually create the CSP (that is, build a data 
  structure).  Feel free to change the signatures of these functions.
  Note that numbering starts from 0: the variables are numbered 0..N-1,
  and the values are numbered 0..K-1.

  INPUT PARAMETERS:
N: number of variables
D: size of each variable's domain
C: number of constraints
T: number of incompatible value pairs in each constraint
Seed: a negative number means start a new sequence of 
pseudo-random numbers; a positive number means continue
with the same sequence.  S is turned positive by ran2().
RETURN VALUE:
Returns 0 if there is a problem; 1 for normal completion.
 *********************************************************************/

int MakeURBCSP(int N, int D, int C, int T, long *Seed, int s, list < pair <int,int> > &edges, stringstream& xmlOutput)
{
	int PossibleCTs, PossibleNGs;       /* CT means "constraint" */
	unsigned long *CTarray, *NGarray;   /* NG means "nogood pair" */
	long selectedCT, selectedNG;
	int i, c, r, t;
	int var1, var2, val1, val2;
	static int instance;

	/* Check for valid values of N, D, C, and T. */
	if (N < 2)
	{
		cout << "MakeURBCSP: ***Illegal value for N: " << N << "\n";
		return 0;
	}
	if (D < 2)
	{
		cout << "MakeURBCSP: ***Illegal value for D: " << D << "\n";
		return 0;
	}
	if (C < 0 || C > N * (N - 1) / 2)
	{
		cout << "MakeURBCSP: ***Illegal value for C: " << C <<"\n";
		return 0;
	}
	if (T < 1 || T > ((D * D) - 1))
	{
		cout << "MakeURBCSP: ***Illegal value for T: " << T << "\n";
		return 0;
	}

	if (*Seed < 0)      /* starting a new sequence of random numbers */
		instance = 0;
	else
		++instance;       /* increment static variable */

	StartCSP(N, D, instance, C, xmlOutput);

	/* The program has to choose randomly and uniformly m values from
	   n possibilities.  It uses the following logic for both constraints
	   and nogood value pairs:
	   1. Let t[] be an array of the n possibilities
	   2. for i = 0 to m-1
	   3.    r = random(i, n-1)    ; random() returns an int in [i,n-1]
	   4.    swap t[i] and t[r]
	   5. end-for
	   At the end of the for loop, the elements from t[0] to t[m-1] are
	   the m randomly selected elements.
	   */

	/* Create an array for each possible binary constraint. */
	PossibleCTs = N * (N - 1) / 2;
	CTarray = new unsigned long[PossibleCTs];

	/* Create an array for each possible value pair. */
	PossibleNGs = D * D;
	NGarray = new unsigned long[PossibleNGs];

	/* Initialize the CTarray.  Each entry has one var in the high two
	   bytes, and the other in the low two bytes. */
	i=0;
	for (var1=0; var1<(N-1); ++var1)
		for (var2=var1+1; var2<N; ++var2)
			CTarray[i++] = (var1 << 16) | var2;

	/* Select C constraints. */
	for (c=0; c<C; ++c)
	{
		/* Choose a random number between c and PossibleCTs - 1, inclusive. */
		r =  c + (int) (ran2(Seed) * (PossibleCTs - c)); 

		/* Swap elements [c] and [r]. */
		selectedCT = CTarray[r];
		CTarray[r] = CTarray[c];
		CTarray[c] = selectedCT;

		/* Broadcast the constraint. */
		AddConstraint(c, (int)(CTarray[c] >> 16), (int)(CTarray[c] & 0x0000FFFF), xmlOutput);

		// Add this edge to the list of edges
		int var1 = (int)(CTarray[c] >> 16);
		int var2 = (int)(CTarray[c] & 0x0000FFFF);
		edges.push_back(pair <int,int>());
		edges.back().first = var1; // add these vars into the newest element
		edges.back().second = var2;		

		/* For each constraint, select T illegal value pairs. */

		/* Initialize the NGarray. */
		for (i=0; i<(D*D); ++i)
			NGarray[i] = i;

		addRelHead ( c, T, s, xmlOutput);

		/* Select T incompatible pairs. */
		for (t=0; t<T; ++t)
		{
			/* Choose a random number between t and PossibleNGs - 1, inclusive.*/
			r =  t + (int) (ran2(Seed) * (PossibleNGs - t));
			selectedNG = NGarray[r];
			NGarray[r] = NGarray[t];
			NGarray[t] = selectedNG;

			/* Broadcast the nogood value pair. */
			AddNogood((int)(NGarray[t] / D), (int)(NGarray[t] % D), xmlOutput);
			if ( t+1<T )
				addTupleEnd(xmlOutput);
		}
		addRelTail(xmlOutput);
	}

	EndCSP(xmlOutput);
	delete[] CTarray;
	delete[] NGarray;
	return 1;
}



/*********************************************************************
  3. This random number generator is from William H. Press, et al.,
  _Numerical Recipes in C_, Second Ed. with corrections (1994), 
  p. 282.  This excellent book is available through the
  WWW at http://nr.harvard.edu/nr/bookc.html.
  The specific section concerning ran2, Section 7.1, is in
http://cfatab.harvard.edu/nr/bookc/c7-1.ps
 *********************************************************************/

#define IM1   2147483563
#define IM2   2147483399
#define AM    (1.0/IM1)
#define IMM1  (IM1-1)
#define IA1   40014
#define IA2   40692
#define IQ1   53668
#define IQ2   52774
#define IR1   12211
#define IR2   3791
#define NTAB  32
#define NDIV  (1+IMM1/NTAB)
#define EPS   1.2e-7
#define RNMX  (1.0 - EPS)

/* ran2() - Return a random floating point value between 0.0 and
   1.0 exclusive.  If idum is negative, a new series starts (and
   idum is made positive so that subsequent calls using an unchanged
   idum will continue in the same sequence). */

float ran2(long *idum)
{
	int j;
	long k;
	static long idum2 = 123456789;
	static long iy = 0;
	static long iv[NTAB];
	float temp;

	if (*idum <= 0) {                             /* initialize */
		if (-(*idum) < 1)                           /* prevent idum == 0 */
			*idum = 1;
		else
			*idum = -(*idum);                         /* make idum positive */
		idum2 = (*idum);
		for (j = NTAB + 7; j >= 0; j--) {           /* load the shuffle table */
			k = (*idum) / IQ1;
			*idum = IA1 * (*idum - k*IQ1) - k*IR1;
			if (*idum < 0)
				*idum += IM1;
			if (j < NTAB)
				iv[j] = *idum;
		}
		iy = iv[0];
	}

	k = (*idum) / IQ1;
	*idum = IA1 * (*idum - k*IQ1) - k*IR1;
	if (*idum < 0)
		*idum += IM1;
	k = idum2/IQ2;
	idum2 = IA2 * (idum2 - k*IQ2) - k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j = iy / NDIV;
	iy = iv[j] - idum2;
	iv[j] = *idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp = AM * iy) > RNMX)
		return RNMX;                                /* avoid endpoint */
	else
		return temp;
}


/*********************************************************************
  4. An implementation of StartCSP, AddConstraint, AddNogood, and EndCSP 
  which prints out the CSP, just listing incompatible value pairs.
  Each constraint starts one a new line, and the id-numbers of the
  variables appear before the colon.  For instance, the output of
  urbcsp 10 5 4 3 9999 10
  begins
  Instance 0
  8   9: (1 1) (4 0) (0 4) 
  2   4: (0 3) (3 1) (4 0) 
  6   9: (4 1) (2 0) (0 3) 
  1   5: (0 3) (4 0) (0 0) 
 *********************************************************************/

void StartCSP(int N, int D, int instance, int C, stringstream &xmlOutput)
{
	if ( order != 0 ) {
		if ( order == 1 ) {
			xmlOutput << "\n<constraints nbConstraints=\"" << C << "\">";
		}
		return;
	}
	int i = 0;	
	//  printf("\nInstance %d", instance);
	xmlOutput << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	xmlOutput << "\n<instance> ";
	xmlOutput << "\n<presentation name=\"Instance" << instance << "\" maxConstraintArity=\"2\" format=\"XCSP 2.0\"/>";

	xmlOutput << "\n<domains nbDomains=\"1\">";
	xmlOutput << "\n<domain name=\"D0\" nbValues=\"" << D << "\">0.." << D-1 << "</domain>";
	xmlOutput << "\n</domains>";
	xmlOutput << "\n";

	xmlOutput << "\n<variables nbVariables=\"" << N << "\">";

	for ( i = 0; i < N; i++ ) {
		xmlOutput << "\n<variable name=\"V" << i << "\" domain=\"D0\"/>"; 
	}

	xmlOutput << "\n</variables>"; 
	xmlOutput << "\n<relations nbRelations=\"" << C << "\">";
}

void AddConstraint(int index, int var1, int var2, stringstream &xmlOutput)
{
	if ( order != 1  ) {
		return;
	}
	xmlOutput << "\n<constraint name=\"C" << index << "\" arity=\"2\" scope=\"V" <<
	 var1 << " V" << var2 << "\" reference=\"R" << index << "\"";
}

void addRelHead ( int index, int t, int s, stringstream &xmlOutput) {
  if ( order == 0 ) {
    if ( s ) {
      xmlOutput << "\n<relation name=\"R" << index
		 << "\" arity=\"2\" nbTuples=\"" << t << "\" semantics=\"supports\">";
    }
    else {
      xmlOutput << "\n<relation name=\"R" << index
		 << "\" arity=\"2\" nbTuples=\"" << t << "\" semantics=\"conflicts\">";
    }
  }
}


void AddNogood(int val1, int val2, stringstream &xmlOutput)
{
	if ( order != 0 )
		return;
	xmlOutput << val1 << " " << val2;
}

void EndCSP(stringstream &xmlOutput)
{
	if ( order == 0 )
		xmlOutput << "\n</relations>";
	else 
		xmlOutput << "\n</constraints>\n</instance>";
	
}

void addTupleEnd(stringstream &xmlOutput) {
	if ( order == 0 )
		xmlOutput << "|";
}

void addRelTail (stringstream &xmlOutput) {

	if ( order == 0 ) 
		xmlOutput << "</relation>";

	if ( order == 1 ) 
		xmlOutput << "></constraint>";
}

// Used to check connectedness of a graph.  See below.
struct VarNode
{
	bool Open;
	list <int> NeighborIDs;
};

bool CheckConnectedness(int numVariables, list <pair <int,int> > &edges)
{
	// This is just silly.  And invalid.
	if (numVariables == 0)
		return false;

	// Indexed by variable ID, naturally
	map<int, VarNode> nodes;

	// We know the variable IDs are 0..N-1, so we can init the nodes:
	for (int i = 0; i < numVariables; i++)
	{
		VarNode temp;
		temp.Open = true;

		nodes[i] = temp;
	}

	// We establish all the relations between variables.
	for (list <pair <int,int> >::iterator i = edges.begin(); i != edges.end(); i++)
	{
		// Connect each node to the other
		nodes[i->first].NeighborIDs.push_back(i->second);
		nodes[i->second].NeighborIDs.push_back(i->first);
	}

	// Now, we visit every node, starting from node 0.
	list <int> openList;
	openList.push_back(0);

	for (list <int>::iterator i = openList.begin(); i != openList.end(); i++)
	{
		// Close the node, now that it has been visited.
		nodes[*i].Open = false;

		list <int> &NborList = nodes[*i].NeighborIDs; // For brevity; see loop

		// Add this node's neighbors to the list, if they're still open.
		for (list <int>::iterator neighbor = NborList.begin(); neighbor != NborList.end(); neighbor++)
		{
			if (nodes[*neighbor].Open)
				openList.push_back(*neighbor);
		}
	}

	// Now we're ready to check connectedness.
	// Check every variable and see if it was visited.
	bool connected = true;
	for (map<int, VarNode>::iterator i = nodes.begin(); i != nodes.end(); i++)
		if (i->second.Open)
		{
			connected = false;
			break;
		}

	return connected;
}
